//: $CC -DLS_LIB -c $file
#ifndef LS_GSC
#define LS_GSC
#include "ls_names.h"
#include LS_GS_H

ls_GS::ls_GS(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); rc=-1;
  A_name=A.name; F_name=F.name; R_name=R.name;
  strcpy(F.name,"Gram-Schmidt-QR-Factorization");
  strcpy(R.name,"upper-triangual-matrix");
}

ls_GS::ls_GS(ls_GS &B)
{ eps=B.eps; strcpy(name=ONAME,B.name); rc=B.rc;
  A_name=A.name; F_name=F.name; R_name=R.name;
}

ls_GS& ls_GS::swap(ls_GS &B)
{ char na[ls_len];strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); R.swap(B.R);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_GS& ls_GS::operator=(const ls_GS &B)
{ R=B.R; rc=B.rc; eps=B.eps; A=B.A; F=B.F;
  return *this;
}

ls_GS::ls_GS(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_GS::ls_GS";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; R_name=R.name;
  ls_UINT i, j, k, dim=A.dimension(), n=A.number_of_columns(),
       m=A.number_of_rows();
  if(m<n) throw ls_error(q1,name,size_error_);
  ls_arrayU<ls_REAL> kR(n,"upper-triangular-matrix");
  R.swap(kR);
  F=A; eps=ls_EPS;
  if(alpha>0)
  {  for (i=0; i<n; F[i][i]+=alpha, i++);
     strcpy(F_name,"regularized_GS-Factorization");
  }
  else strcpy(F_name,"GS-Factorization");
  double s;
  ls_REAL *aa=F.asArray(), *ae=aa+dim;

  for(k=1; k<n; k++)
  { for(s=0, i=0; i<m; s+=F[i][k]*F[i][k], i++);
    if((s=sqrt(s))<eps){ rc=1+k; F.remove(); R.remove(); return;}
    R[k][k]=s;
    for(i=0, s=1/s; i<m; F[i][k]=F[i][k]*s, i++);
    for(j=k+1; j<n; j++)
    { for(s=0, i=0; i<m; s+=F[i][k]*F[i][j], i++);
      R[j][k]=s;
      for(i=0; i<m; F[i][j]-=F[i][k]*s, i++);
    }
  }
  rc=0;
}

const ls_GS& ls_GS::solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_GS::solve";
  ls_UINT i, k, nn=b.dimension(), n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(nn!=m) throw ls_error(q1,b.name,size_error_);
  ls_Vector z; z=b*F;
  ls_REAL s,sp,sm,*aa,*za=z.asArray(),*ze=za+m, *xx=x.asArray();
  for (k=n, aa=R[k-1]; k--; xx[k]=(za[k]-(sp+sm))/R[k][k], aa-=n)
    for (i=k+1, sp=sm=0; i<n; i++)
      if((s=aa[i]*xx[i])<0) sm+=s; else sp+=s;
  return *this;
}

ls_UINT ls_GS::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_GS::post_iteration";
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(sizeof(ls_REAL)<8) throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_GS::residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_GS::Qx(const ls_Vector &x) const
{ static char *q1="ls_GS::Qx";
  ls_UINT n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  return F*x;
}

ls_Vector ls_GS::xQ(const ls_Vector &x) const
{ static char *q1="ls_GS::xQ";
  ls_UINT n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  return x*F;
}

ls_Vector ls_GS::Rx(const ls_Vector &x) const
{ static char *q1="ls_GS::Rx";
  ls_UINT i,j,n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  ls_Vector v(n);
  ls_REAL s,*xx=x.asArray(),*vv=v.asArray(),*aa;
  for(i=0; i<n; vv[i]=s, i++)
    for(j=i, aa=R[i], s=0; j<n; s+=*(aa++)*xx[j], j++);
  return v;
}

ls_Vector ls_GS::xR(const ls_Vector &x) const
{ static char *q1="ls_GS::xR";
  ls_UINT i,j,n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  ls_Vector v(n);
  ls_REAL s,*xx=x.asArray(),*vv=v.asArray();
  for(j=0; j<n; vv[j]=s, j++)
    for(i=0, s=0; i<j; s+=R[i][j]*xx[i], i++);
  return v;
}

const ls_GS& ls_GS::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f); R.write_row(f);
  f<<"}"<<endl;
  return *this;
}

ls_GS& ls_GS::read_row(istream &f)
{ static char *q1="ls_GS::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(R.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); R.read_row(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_GS& ls_GS::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); R.write_column(f);
  f<<"}"<<endl;
  return *this;
}

ls_GS& ls_GS::read_column(istream &f)
{ static char *q1="ls_LU_column::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(R.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); R.read_column(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_GS& ls_GS::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
  }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_GS& ls_GS::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
