//: $CC -DLS_LIB -c $file
#ifndef LS_INVC
#define LS_INVC
#include "ls_names.h"
#include LS_INV_H

ls_INV::ls_INV(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name;
  strcpy(F_name,"Inversion"); rc=-1;
}

ls_INV::ls_INV(ls_INV &B)
{ eps=B.eps; rc=B.rc; strcpy(name=ONAME,B.name);
  A_name=A.name; F_name=F.name;
}

ls_INV& ls_INV::swap(ls_INV &B)
{ char na[ls_len];strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_INV& ls_INV::operator=(const ls_INV &B)
{ eps=B.eps; A=B.A; F=B.F;rc=B.rc; return *this;}

ls_INV::ls_INV(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_INV::ls_INV";
  A.swap(B); strcpy(name=ONAME, nam); F_name=F.name; A_name=A.name;
  ls_UINT i, n=A.number_of_columns();
  if(n!=A.number_of_rows())
    throw ls_error(A_name,q1,object_not_quadratic_);
  F=A; eps=ls_EPS;
  double piv, s;
  ls_REAL *ai, *aai, *ak, *aak, *a=F.asArray(), *ae=a+n*n, *aae;
  if(alpha>0)
  { for(ai=a, i=0; i<n; ai[i]+=alpha, i++, ai+=n);
    strcpy(F_name,"regularized_Inversion");
  }
  else strcpy(F_name,"Inversion");
  for(i=0, ai=a; i<n; i++, ai+=n)
  { if(fabs(piv=double(ai[i]))<eps){ rc=i+1; F.remove(); return;}
    for(piv=-1/piv, ai[i]=-1, aai=ai, aae=ai+n; aai<aae; *aai*=ls_REAL(piv), aai++);
    for(ak=a; ak<ae; ak+=n)
    { if((fabs(s=double(ak[i]))<eps)||(ak==ai)) continue;
      ak[i]=0; for(aai=ai, aak=ak; aai<aae; *aak+=ls_REAL(*aai*s), aai++, aak++);
    }
  }
  rc=0;
}

const ls_INV& ls_INV::solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_INV::solve";
  if(rc) throw ls_error(name,q1,object_is_empty_);
  x=F*b; return *this;
}

ls_UINT ls_INV::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_INV::post_iteration";
  if(sizeof(ls_REAL)<8) throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,object_is_empty_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_INV::residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_INV::Fx(const ls_Vector &x) const
{ static char *q1="ls_INV::Fx";
  ls_UINT n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,object_is_empty_);
  return F*x;
}

ls_Vector ls_INV::xF(const ls_Vector &x) const
{ static char *q1="ls_INV::xF";
  ls_UINT n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,object_is_empty_);
  return x*F;
}

const ls_INV& ls_INV::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f);
  f<<"}"<<endl;
  return *this;
}

ls_INV& ls_INV::read_row(istream &f)
{ static char *q1="ls_INV::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_INV& ls_INV::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f);
  f<<"}"<<endl;
  return *this;
}

ls_INV& ls_INV::read_column(istream &f)
{ static char *q1="ls_INV::read_column";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_INV& ls_INV::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_INV& ls_INV::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
