//: $CC -DLS_LIB -c $file
#ifndef LS_INV_DIAGONALC
#define LS_INV_DIAGONALC
#include "ls_names.h"
#include LS_INVDIAGONAL_H

ls_INV_diagonal::ls_INV_diagonal(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name; ind_name=ind.name;
  strcpy(F_name,"Inversion"); rc=-1;
}

ls_INV_diagonal::ls_INV_diagonal(ls_INV_diagonal &B)
{ eps=B.eps; rc=B.rc; strcpy(name=ONAME,B.name);
  A_name=A.name; F_name=F.name; ind_name=ind.name;
}

ls_INV_diagonal& ls_INV_diagonal::swap(ls_INV_diagonal &B)
{ char na[ls_len];strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); ind.swap(B.ind);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_INV_diagonal& ls_INV_diagonal::operator=(const ls_INV_diagonal &B)
{ eps=B.eps; A=B.A; F=B.F; ind=B.ind; rc=B.rc; return *this;}

ls_INV_diagonal::ls_INV_diagonal(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_INV_diagonal::ls_INV_diagonal";
  A.swap(B); strcpy(name=ONAME, nam); F_name=F.name; A_name=A.name;
  ls_UINT i, j, n=A.number_of_columns();
  if(n!=A.number_of_rows())
    throw ls_error(name,q1,object_not_quadratic_);
  ls_array1<ls_UINT> kind(n,"Diagonal-Indices"); ind.swap(kind);
  ls_UINT *in=ind.asArray(), *jj, js, *kk, *nc=in+n;
  F=A; eps=ls_EPS;
  ls_Vector col(n), row(n);
  double piv, s;
  ls_REAL *ai, *a=F.asArray(), *ak, *aai, *aak, *aae,
          *cc=col.asArray(), *rr=row.asArray();
  if(alpha>0)
  { for(ai=a, i=0; i<n; ai[i]+=alpha, i++, ai+=n);
    strcpy(F_name,"regularized_diagonal_Inversion");
  }
  else strcpy(F_name,"diagonal_Inversion");
  for (ai=a, i=0; i<n; ind[i]=i, rr[i]=(ls_REAL)s, i++, ai+=n)
    for(s=0, j=0; j<n; piv=fabs(double(ai[j])), s+=piv, cc[j]+=(ls_REAL)piv, j++);
  for (i=0; i<n; rr[i]*=cc[i], i++);
  for(i=0; i<n; i++)
  { for(jj=in+i,piv=fabs(F[*jj][*jj]*row[*jj]), kk=jj;++kk<nc;)
      if((s=fabs(F[*kk][*kk]*row[*kk]))>piv) jj=kk, piv=s;
    if(piv<eps){ rc=i+1; F.remove(); ind.remove(); return;}
    js=*jj, *jj=in[i], in[i]=js; ai=F[js];
    for(piv=-1/ai[js], ai[js]=-1, aai=ai, aae=aai+n; aai<aae; *aai*=(ls_REAL)piv, aai++);
    for(kk=in; kk<nc; kk++)
    { ak=F[*kk]; if((fabs(s=double(ak[js]))<eps)||(ak==ai)) continue;
      ak[js]=0; for(aak=ak, aai=ai; aai<aae; *aak+=ls_REAL(*aai*s), aai++, aak++);
    }
  }
  rc=0;
}

const ls_INV_diagonal& ls_INV_diagonal::solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_INV_diagonal::solve";
  if(rc) throw ls_error(name,q1,object_is_empty_);
  x=Fx(b); return *this;
}

ls_UINT ls_INV_diagonal::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_INV_diagonal::post_iteration";
  if(sizeof(ls_REAL)<8) throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,object_is_empty_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_INV_diagonal::residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_INV_diagonal::Fx(const ls_Vector &x) const
{ static char *q1="ls_INV_diagonal::Fx";
  ls_UINT *jj, n=F.number_of_columns(), *in=ind.asArray(),
          *ii, *nn=in+n;
  if(!n) throw ls_error(name,q1,object_is_empty_);
  ls_Vector v(n);
  ls_REAL sp, sm, s, *xx=x.asArray(),*vv=v.asArray(), *ai;
  for(ii=in; ii<nn; v[*ii]=sp+sm, ii++)
    for(jj=in, ai=F[*ii], sp=sm=0; jj<nn; jj++)
      if((s=ai[*jj]*xx[*jj])<0) sm+=s; else sp+=s;
  return v;
}

ls_Vector ls_INV_diagonal::xF(const ls_Vector &x) const
{ static char *q1="ls_INV_diagonal::xF";
  ls_UINT *ii, n=F.number_of_columns(), *in=ind.asArray(),
          *jj, *nn=in+n;
  if(!n) throw ls_error(name,q1,object_is_empty_);
  ls_Vector v(n);
  ls_REAL sp, sm, s, *xx=x.asArray(), *vv=v.asArray();
  for(jj=in; jj<nn; vv[*jj]=sp+sm, jj++)
    for(ii=in, sp=sm=0; ii<nn; ii++)
      if((s=F[*ii][*jj]*xx[*ii])<0) sm+=s; else sp+=s;
  return v;
}

const ls_INV_diagonal& ls_INV_diagonal::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_INV_diagonal& ls_INV_diagonal::read_row(istream &f)
{ static char *q1="ls_INV_diagonal::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_INV_diagonal& ls_INV_diagonal::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_INV_diagonal& ls_INV_diagonal::read_column(istream &f)
{ static char *q1="ls_INV_diagonal::read_column";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_INV_diagonal& ls_INV_diagonal::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_INV_diagonal& ls_INV_diagonal::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
