//: $CC -DLS_LIB -c $file
#ifndef LS_INV_TOTALC
#define LS_INV_TOTALC
#include "ls_names.h"
#include LS_INVTOTAL_H

ls_INV_total::ls_INV_total(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name;
  indr_name=indr.name; indc_name=indc.name;
  strcpy(F_name,"Inversion"); rc=-1;
}

ls_INV_total::ls_INV_total(ls_INV_total &B)
{ eps=B.eps; rc=B.rc; strcpy(name=ONAME,B.name);
  A_name=A.name; F_name=F.name; indr_name=indr.name; indc_name=indc.name;
}

ls_INV_total& ls_INV_total::swap(ls_INV_total &B)
{ char na[ls_len];strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); indr.swap(B.indr); indc.swap(B.indc);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_INV_total& ls_INV_total::operator=(const ls_INV_total &B)
{ eps=B.eps; A=B.A; F=B.F; indr=B.indr; indc=B.indc; rc=B.rc; return *this;}

ls_INV_total::ls_INV_total(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_INV_total::ls_INV_total";
  A.swap(B); strcpy(name=ONAME, nam); F_name=F.name; A_name=A.name;
  ls_UINT i, j, k, n=A.number_of_columns();
  if(n!=A.number_of_rows())
    throw ls_error(name,q1,object_not_quadratic_);
  ls_array1<ls_UINT> kindr(n,"Row-Indices"); indr.swap(kindr);
  ls_array1<ls_UINT> kindc(n,"Column-Indices"); indc.swap(kindc);
  ls_UINT *inr=indr.asArray(), *inc=indc.asArray(),
          *ii, *jj, *ip, *jp, *nr=inr+n, *nc=inc+n, iz, js;
  F=A; eps=ls_EPS;
  ls_Vector col(n), row(n);
  double piv, s;
  ls_REAL *ai, *ak, *a=F.asArray(), *aai, *aak, *aae,
          *rr=row.asArray(), *cc=col.asArray();
  if(alpha>0)
  { for(ai=a, i=0; i<n; ai[i]+=alpha, i++, ai+=n);
    strcpy(F_name,"regularized_total_Inversion");
  }
  else strcpy(F_name,"total_Inversion");
  for (i=0, ai=a; i<n; inr[i]=inc[i]=i, rr[i]=s, i++)
    for(s=0, j=0; j<n; piv=fabs(*ai), s+=piv, cc[j]+=piv, j++, ai++);
  for (i=0; i<n; i++)
  { for(ii=ip=inr+i,jp=inc+i, piv=fabs(F[*ip][*jp]*rr[*ip]*cc[*jp]);++ii<nr;)
      for(jj=inc+i, ai=F[*ii]; jj<nc; jj++)
        if((s=fabs(ai[*jj]*rr[*ii]*cc[*jj]))>piv) piv=s, jp=jj,ip=ii;
    if(piv<eps){ rc=i+1; F.remove(); indr.remove(); indc.remove(); return;}
    iz=*ip, *ip=inr[i], inr[i]=iz;
    js=*jp, *jp=inc[i], inc[i]=js;
    ai=F[iz];
    for(piv=-1/ai[js], ai[js]=-1, aai=ai, aae=ai+n; aai<aae; *aai*=piv, aai++);
    for(k=0; k<n; k++)
    { ak=F[k]; if((fabs(s=double(ak[js]))<eps)||(ak==ai)) continue;
      ak[js]=0; for(aak=ak, aai=ai; aai<aae; *aak+=*aai*s, aai++, aak++);
    }
  }
  rc=0;
}

const ls_INV_total& ls_INV_total::solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_INV_total::solve";
  if(rc) throw ls_error(name,q1,object_is_empty_);
  x=Fx(b); return *this;
}

ls_UINT ls_INV_total::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_INV_total::post_iteration";
  if(sizeof(ls_REAL)<8) throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_UINT rep=10, l=0, n=F.number_of_rows();
  if(!n) throw ls_error(name,q1,object_is_empty_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_INV_total::residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_INV_total::Fx(const ls_Vector &x) const
{ static char *q1="ls_INV_total::Fx";
  ls_UINT *jj, n=F.number_of_rows(), *inr=indr.asArray(),
          *inc=indc.asArray(), *ii, *nr=inr+n, *nc=inc+n;
  if(!n) throw ls_error(name,q1,object_is_empty_);
  ls_Vector v(n);
  ls_REAL sp, sm, s, *xx=x.asArray(), *vv=v.asArray(), *aa;
  for(ii=inr; ii<nr; v[*ii]=sp+sm, ii++)
    for(jj=inc, sp=sm=0, aa=F[*ii]; jj<nc; jj++)
      if((s=aa[*jj]*xx[*jj])<0) sm+=s; else sp+=s;
  return v;
}

ls_Vector ls_INV_total::xF(const ls_Vector &x) const
{ static char *q1="ls_INV_total::xF";
  ls_UINT *ii, n=F.number_of_rows(), *inr=indr.asArray(),
          *inc=indc.asArray(), *jj, *nc=inc+n, *nr=inr+n;
  if(!n) throw ls_error(name,q1,object_is_empty_);
  ls_Vector v(n);
  ls_REAL sp, sm, s, *xx=x.asArray(), *a=F.asArray(),
          *vv=v.asArray();
  for(jj=inc; jj<nc; vv[*jj]=sp+sm, jj++)
    for(ii=inr, sp=sm=0; ii<nr; ii++)
      if((s=F[*ii][*jj]*xx[*ii])<0) sm+=s; else sp+=s;
  return v;
}

const ls_INV_total& ls_INV_total::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f); indr.write(f), indc.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_INV_total& ls_INV_total::read_row(istream &f)
{ static char *q1="ls_INV_total::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); indr.read(f), indc.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_INV_total& ls_INV_total::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); indr.write(f), indc.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_INV_total& ls_INV_total::read_column(istream &f)
{ static char *q1="ls_INV_total::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); indr.read(f), indc.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_INV_total& ls_INV_total::operator>> (char *nam) const
{ if(!*nam) write_row(cout);
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.4)), write_row(f);
  }
  return *this;
}

ls_INV_total& ls_INV_total::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
