//: $CC -DLS_LIB -c $file
#ifndef LS_LDLTC
#define LS_LDLTC
#include "ls_names.h"
#include LS_LDLT_H

ls_LDLT::ls_LDLT(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name; rc=-1;
  strcpy(F.name,"LDLT-Factorization");}

ls_LDLT::ls_LDLT(ls_LDLT &B)
{ eps=B.eps; strcpy(name=ONAME,B.name);
  A_name=A.name; F_name=F.name, rc=B.rc;
}

ls_LDLT& ls_LDLT::swap(ls_LDLT &B)
{ char na[ls_len];
  strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_LDLT& ls_LDLT::operator=(const ls_LDLT &B)
{ eps=B.eps; rc=B.rc; A=B.A; F=B.F; return *this;}


ls_LDLT::ls_LDLT(ls_sMatrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_LDLT::ls_LDLT";
  A.swap(B); strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name;
  ls_UINT i, j, k, n=A.dimension(); eps=ls_EPS;
  F=A; eps=ls_EPS;
  if(alpha>0)
  { for(i=0; i<n; F[i][i]+=alpha, i++);
    strcpy(F_name,"regularized_LDLT-Factorization");
  }
  else strcpy(F_name,"LDLT-Factorization");
  double s, t, e;
  ls_REAL *ak, *ai;
  for (k=0; k<n-1; k++)
  { ak=F[k];
    if(fabs(t=double(ak[k]))<eps){ rc=k+1; F.remove(); return;}
    for (e=ak[k]*eps, t=1./t, i=k+1; i<n; ak[i]=(ls_REAL)s, i++)
      if(fabs(s=ak[i]*t)<e){ ak[i]=0; continue;}
      else for (ai=F[i], j=i; j<n; ai[j]-=ls_REAL(ak[j]*s), j++);
  }
  rc=0;
}

const ls_LDLT& ls_LDLT::\
solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_LDLT::solve";
  ls_UINT k, i, n=F.dimension();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_REAL s, *ak, *xx=x.asArray(); x=b;
  for (k=0; k<n; xx[k]=s/ak[k], k++)
    for (ak=F[k], s=xx[k], i=k+1; i<n; xx[i]-=ak[i]*s, i++);
  for (k=n; k--;)
    for (s=xx[k], i=0; i<k; xx[i]-=F[i][k]*s, i++);
  return *this;
}

ls_UINT ls_LDLT::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_LDLT::post_iteration";
  ls_UINT rep=10, l=0, n=F.dimension();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(sizeof(ls_REAL)<8) 
    throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_LDLT::\
residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_LDLT::Lx(const ls_Vector &x) const
{ static char *q1="ls_LDLT::Lx";
  ls_UINT n=F.dimension(), ll;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n), d(n);
  ls_REAL s, *xa=x.asArray(), *xx, *vv=v.asArray(), 
          *aa=F.asArray(), *cc, *c,
          *da=d.asArray(), *dd, *de=da+n, *d0;
  for(dd=da, cc=aa, xx=xa, ll=n; dd<de; dd++, xx++, cc+=ll, ll--)
  { if(!(*cc>eps)) 
    throw ls_error(name,q1,object_not_positive_definit_);
    *dd=(ls_REAL)(*xx*sqrt(*cc));
  }
  for(dd=da, cc=aa, n--; dd<de; *vv=*dd+s, dd++, vv++, cc++)
    for(d0=da, c=cc, ll=n, s=0; d0<dd; 
        s+=*d0*(*c), d0++, c+=ll, ll--);
  return v;
}

ls_Vector ls_LDLT::xL(const ls_Vector &x) const
{ static char *q1="ls_LDLT::xL";
  ls_UINT n=F.dimension(), ll;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xa=x.asArray(), *xx, *xe=xa+n, *y, *va=v.asArray(), 
          *vv, *ve, *aa=F.asArray(), *cc;
  for(cc=aa, xx=xa, vv=va; xx<xe; *vv=*xx+s, vv++, xx++)
    for(y=xx+1, s=0; y<xe; s+=*y*(*cc), y++, cc++);
  for(ve=va+n, vv=va, cc=aa, ll=n; vv<ve; vv++, cc+=ll, ll--)
  { if(!(*cc>eps)) 
      throw ls_error(name,q1,object_not_positive_definit_);
    *vv*=(ls_REAL)sqrt(*cc);
  }
  return v;
}

const ls_LDLT& ls_LDLT::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f);
  f<<"}"<<endl;
  return *this;
}

ls_LDLT& ls_LDLT::read_row(istream &f)
{ static char *q1="ls_LDLT::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LDLT& ls_LDLT::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f);
  f<<"}"<<endl;
  return *this;
}

ls_LDLT& ls_LDLT::read_column(istream &f)
{ static char *q1="ls_LDLT::read_column";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LDLT& ls_LDLT::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_LDLT& ls_LDLT::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
