//: $CC -DLS_LIB -c $file
#ifndef LS_LDLT_DIAGONALC
#define LS_LDLT_DIAGONALC
#include "ls_names.h"
#include LS_LDLTDIAGONAL_H

ls_LDLT_diagonal::ls_LDLT_diagonal(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); rc=-1;
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  strcpy(F.name,"LDLT_diagonal-Factorization");
  strcpy(ind.name,"Diagonal-Indices");
}

ls_LDLT_diagonal::ls_LDLT_diagonal(ls_LDLT_diagonal &B)
{ eps=B.eps; strcpy(name=ONAME,B.name); rc=B.rc;
  A_name=A.name; F_name=F.name; ind_name=ind.name;
}

ls_LDLT_diagonal& ls_LDLT_diagonal::swap(ls_LDLT_diagonal &B)
{ char na[ls_len];
  strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); ind.swap(B.ind);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_LDLT_diagonal& ls_LDLT_diagonal::\
operator=(const ls_LDLT_diagonal &B)
{ eps=B.eps; rc=B.rc; ind=B.ind; A=B.A; F=B.F; return *this;}

ls_LDLT_diagonal::ls_LDLT_diagonal(ls_sMatrix &B, 
                                   ls_REAL alpha, char *nam)
{ static char *q1="ls_LDLT_diagonal::ls_LDLT_diagonal";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  ls_UINT k, i, j, n=A.dimension();
  ls_array1<ls_UINT> kind(n,"Diagonal-Indices"); ind.swap(kind);
  F=A;
  if(alpha>0)
  { for(i=0; i<n; F[i][i]+=alpha, i++);
    strcpy(F_name,"regularized_LDLT_diagonal-Factorization");
  }
  else strcpy(F_name,"LDLT_diagonal-Factorization");
  eps=ls_EPS;
  ls_UINT *in=ind.asArray();
  double s, t, piv;
  ls_REAL *ak, *ai;
  for (k=0; k<n; in[k]=k, k++);
  for (k=0; k<n-1; k++)
  { ak=F[k];
    if(fabs(piv=double(ak[k]))<eps)
    { rc=k+1; F.remove(); ind.remove();return;}
    for (t=fabs(piv), j=k, i=k+1; i<n; i++)
      if(fabs(s=double(F[i][i]))>t) t=s, j=i;
    if(k<j)
      F.swap(k,j),i=in[k],in[k]=in[j],in[j]=i,piv=ak[k];
    for (piv=1./piv, i=k+1; i<n; ak[i]=(ls_REAL)s, i++)
      for (s=ak[i]*piv, ai=F[i], j=i; j<n; 
           ai[j]-=ls_REAL(ak[j]*s), j++);
  }
  rc=0;
}

const ls_LDLT_diagonal& ls_LDLT_diagonal::\
solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_LDLT_diagonal::solve";
  ls_UINT k, i, *in=ind.asArray(), n=F.dimension();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_REAL s, *ak, *xx=x.asArray(); x=b;
  for (k=0; k<n; xx[in[k]]=s/ak[k], k++)
  { ak=F[k];
    for(s=xx[in[k]], i=k+1; i<n; xx[in[i]]-=ak[i]*s, i++);
  }
  for (k=n; k--;)
    for (s=xx[in[k]], i=0; i<k; xx[in[i]]-=F[i][k]*s, i++);
  return *this;
}

ls_UINT ls_LDLT_diagonal::\
post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_LDLT_diagonal::post_iteration";
  ls_UINT rep=10, l=0, n=F.dimension();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(sizeof(ls_REAL)<8) 
    throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_LDLT_diagonal::\
residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_LDLT_diagonal::Lx(const ls_Vector &x) const
{ static char *q1="ls_LDLT_diagonal::Lx";
  ls_UINT *ii, *jj, n=F.dimension(), *in=ind.asArray(), *nn=in+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n), d(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa=F.asArray(),
          *dd=d.asArray();
  for(ii=in; ii<nn; ii++)
  { if(!((s=F[*ii][*ii])>eps))
      throw ls_error(name,q1,object_not_positive_definit_);
    dd[*ii]=(ls_REAL)(xx[*ii]*sqrt(s));
	 }
  for(jj=in; jj<nn; vv[*jj]=dd[*jj]+s, jj++)
    for(ii=in, s=0; ii<jj; s+=dd[*ii]*F[*ii][*jj], ii++);
  return v;
}

ls_Vector ls_LDLT_diagonal::xL(const ls_Vector &x) const
{ static char *q1="ls_LDLT_diagonal::xL";
  ls_UINT *ii, *jj, n=F.dimension(), *in=ind.asArray(), *nn=in+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa;
  for(ii=in; ii<nn; vv[*ii]=xx[*ii]+s, ii++)
    for(jj=ii+1, s=0, aa=F[*ii]; jj<nn; s+=xx[*jj]*aa[*jj], jj++);
  for(ii=in; ii<nn; ii++)
  { if(!((s=F[*ii][*ii])>eps))
      throw ls_error(name,q1,object_not_positive_definit_);
    xx[*ii]*=(ls_REAL)sqrt(s);
  }
  return v;
}

const ls_LDLT_diagonal& ls_LDLT_diagonal::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LDLT_diagonal& ls_LDLT_diagonal::read_row(istream &f)
{ static char *q1="ls_LDLT_diagonal::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LDLT_diagonal& ls_LDLT_diagonal::\
write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LDLT_diagonal& ls_LDLT_diagonal::read_column(istream &f)
{ static char *q1="ls_LDLT_diagonal::read_column";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LDLT_diagonal& ls_LDLT_diagonal::\
operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_LDLT_diagonal& ls_LDLT_diagonal::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
