//: $CC -DLS_LIB -c $file
#ifndef LS_LUC
#define LS_LUC
#include "ls_names.h"
#include LS_LU_H

ls_LU::ls_LU(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name;
  strcpy(F_name,"LU-Factorization"); rc=-1;
}

ls_LU::ls_LU(ls_LU &B)
{ eps=B.eps; rc=B.rc; strcpy(name=ONAME,B.name);
  A_name=A.name; F_name=F.name;
}

ls_LU& ls_LU::swap(ls_LU &B)
{ char na[ls_len];
  strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_LU& ls_LU::operator=(const ls_LU &B)
{ eps=B.eps; A=B.A; F=B.F;rc=B.rc; return *this;}

ls_LU::ls_LU(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_LU::ls_LU";
  A.swap(B); strcpy(name=ONAME, nam); F_name=F.name; A_name=A.name;
  ls_UINT j, m=A.number_of_rows(), n=A.number_of_columns(), nn;
  if(m!=n) throw ls_error(A_name,q1,object_not_quadratic_);
  F=A; eps=ls_EPS;
  if(alpha>0)
  { for(ls_UINT i=0; i<n; F[i][i]+=alpha, i++);
    strcpy(F_name,"regularized_LU-Factorization");
  }
  else strcpy(F_name,"LU-Factorization");
  double piv, s, e;
  ls_REAL *ak, *ai, *ae;
  for (ai=F.asArray(), ae=ai+n*n, nn=n+1; --nn; ai+=(n+1))
  { if(fabs(double(*ai))<eps){ rc=n-nn+1; F.remove(); return;}
    for (e=*ai*eps, piv=1/(*ai), ak=ai+n; ak<ae; ak+=n)
    { if(fabs(s=*ak*piv)<e){ *ak=0; continue;}
      for (*ak=(ls_REAL)s, j=0; ++j<nn; ak[j]-=ls_REAL(ai[j]*s));
    }
  }
  rc=0;
}

const ls_LU& ls_LU::solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_LU::solve";
  ls_UINT i, j, dim=F.dimension(), n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_REAL s,sp,sm,*aa,*xx=x.asArray(),*bb=b.asArray(),*a=F.asArray();
  for (i=0, aa=a; i<n; xx[j]=*bb-(sp+sm), i++, aa+=n, bb++)
    for (j=0, sp=sm=0; j<i; j++)
      if((s=aa[j]*xx[j])<0) sm+=s; else sp+=s;
  for (i=n, aa=a, aa+=(dim-n); i--; 
       xx[i]=(xx[i]-(sp+sm))/aa[i], aa-=n)
    for (j=i, sp=sm=0; ++j<n;)
      if((s=aa[j]*xx[j])<0) sm+=s; else sp+=s;
  return *this;
}

ls_UINT ls_LU::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_LU::post_iteration";
  if(sizeof(ls_REAL)<8) 
    throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_LU::\
residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_LU::Lx(const ls_Vector &x) const
{ static char *q1="ls_LU::Lx";
  ls_UINT i, j, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa=F.asArray();
  for(i=0; i<n; vv[i]=xx[i]+s, i++, aa+=n)
    for(j=0, s=0; j<i; s+=xx[j]*aa[j], j++);
  return v;
}

ls_Vector ls_LU::xL(const ls_Vector &x) const
{ static char *q1="ls_LU::xL";
  ls_UINT i, j, n=F.number_of_columns(), nn=n-1;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s,*xx=x.asArray(),*vv=v.asArray(),*aa;
  for(j=0; j<nn; vv[j]=xx[j]+s, j++)
    for(i=j+1, s=0, aa=&F[i][j]; i<n; s+=*aa*xx[i], i++, aa+=n);
  vv[nn]=xx[nn];
  return v;
}

ls_Vector ls_LU::Ux(const ls_Vector &x) const
{ static char *q1="ls_LU::Ux";
  ls_UINT i, j, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa=F.asArray();
  for(i=0; i<n; vv[i]=s, i++, aa+=n)
    for(j=i, s=0; j<n; s+=xx[j]*aa[j], j++);
  return v;
}

ls_Vector ls_LU::xU(const ls_Vector &x) const
{ static char *q1="ls_LU::xU";
  ls_UINT i, j, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa=F.asArray(), *cc;
  for(j=0; j<n; vv[j]=*cc*xx[j]+s, j++)
    for(i=0, s=0, cc=&aa[j]; i<j; s+=*cc*xx[i], i++, cc+=n);
  return v;
}

const ls_LU& ls_LU::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU& ls_LU::read_row(istream &f)
{ static char *q1="ls_LU::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU& ls_LU::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU& ls_LU::read_column(istream &f)
{ static char *q1="ls_LU::read_column";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU& ls_LU::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_LU& ls_LU::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
