#ifndef LS_LU
#define LS_LU
#include "ls_names.h"
#include LS_MATRIX_H

class ls_LU
{ protected:
  ls_Matrix A, F;        // Matrix und Faktorisierung
  char ONAME[ls_len];
  public:
  char *name, *A_name, *F_name;
  ls_REAL eps;           // Genauigkeitsschranke
  int rc;                // Rckkehrwert nach Faktorisierung
  ls_LU(char* ="ls_LU");
  ls_LU(ls_Matrix &, ls_REAL = 0., char* ="ls_LU");
                         // Matrix-bernahme (move-Konstruktor)
                         // LU-Faktorisierung ohne Pivotisierung
  ls_LU(ls_LU &);        // move-Konstruktor
  unsigned char good()const{ return !rc;} // Erfolgssignal
  ls_LU& swap(ls_LU &);
  const ls_LU& operator=(const ls_LU &);
  const ls_LU& solve(ls_Vector &x, const ls_Vector &b) const;
                         // Gleichungslser
  ls_UINT post_iteration(ls_Vector &x, ls_Vector &b) const;
                         // Nachiteration
  ls_Vector residuum(const ls_Vector &x, 
                     const ls_Vector &b) const;
  ls_Vector Lx(const ls_Vector &x) const; // L*x
  ls_Vector xL(const ls_Vector &x) const; // x*L
  ls_Vector Ux(const ls_Vector &x) const; // U*x
  ls_Vector xU(const ls_Vector &x) const; // x*U
  const ls_LU& write_row(ostream &) const;
  ls_LU& read_row(istream &);
  const ls_LU& write_column(ostream &) const;
  ls_LU& read_column(istream &);
  const ls_LU& operator>> (char *) const;
  ls_LU& operator<< (char *);
};
#ifndef LS_LIB
#include LS_LU_C
#endif
#endif
