//: $CC -DLS_LIB -c $file
#ifndef LS_LU_COLUMNC
#define LS_LU_COLUMNC
#include "ls_names.h"
#include LS_LUCOLUMN_H

ls_LU_column::ls_LU_column(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); rc=-1;
  strcpy(F.name,"ls_LU_column-Factorization");
  strcpy(ind.name,"Column-Indices");
  A_name=A.name; F_name=F.name; ind_name=ind.name;
}

ls_LU_column::ls_LU_column(ls_LU_column &B)
{ eps=B.eps; strcpy(name=ONAME,B.name); rc=B.rc;
  A_name=A.name; F_name=F.name; ind_name=ind.name;
}

ls_LU_column& ls_LU_column::swap(ls_LU_column &B)
{ char na[ls_len];
  strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); ind.swap(B.ind);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_LU_column& ls_LU_column::operator=(const ls_LU_column &B)
{ eps=B.eps; rc=B.rc; ind=B.ind; A=B.A; F=B.F; return *this;}

ls_LU_column::ls_LU_column(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_LU_column::ls_LU_column";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  ls_UINT n=A.number_of_columns(), m=A.number_of_rows(),
          dim=A.dimension();
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  ls_array1<ls_UINT> kind(n,"Column-Indices"); ind.swap(kind);
  ls_UINT j,*in=ind.asArray(), n1=n, *ii, *jj, *js, *nn=in+n;
  F=A; eps=ls_EPS;
  if(alpha>0)
  { for(ls_UINT i=0; i<n; F[i][i]+=alpha, i++);
    strcpy(F_name,"regularized_LU_column-Factorization");
  }
  else strcpy(F_name,"LU_column-Factorization");
  ls_Vector x(n);
  double piv, s, e;
  ls_REAL *a=F.asArray(), *ak, *ai, *ae=a+dim, *xx=x.asArray();
  for (j=0; j<n; in[j]=j, xx[j]=(ls_REAL)(1./s), j++)
    for (s=0, ai=a+j; ai<ae; s+=fabs(double(*ai)), ai+=n);
  for (ii=in, ai=a; --n1; ai+=n, ii++)
  { for(piv=fabs(ai[*ii]*xx[*ii]), js=jj=ii; ++jj<nn;)
     if((s=fabs(ai[*jj]*xx[*jj]))>piv) piv=s, js=jj;
    if(piv<eps)
    { rc=1+(ls_UINT)(ii-in); F.remove(); ind.remove(); return;}
    m=*ii, *ii=*js, *js=m;
    for (piv=double(ai[*ii]), e=piv*eps, piv=1/piv, ak=ai+n;
         ak<ae; ak+=n)
    { if(fabs(s=piv*ak[*ii])<e){ ak[*ii]=0; continue;}
      for(ak[*ii]=(ls_REAL)s, jj=ii; ++jj<nn; 
        ak[*jj]-=ls_REAL(ai[*jj]*s));
    }
  }
  rc=0;
}

const ls_LU_column& ls_LU_column::\
solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_LU_column::solve";
  ls_UINT *in=ind.asArray(), n=F.number_of_columns(),
          *jj, *ii, *nn=in+n;
  if(!n) throw ls_error("ls_LU_column",q1,no_factorization_given_);
  ls_REAL s, sp, sm, *xx=x.asArray(), *bb=b.asArray(), *ai;
  for (ii=in, ai=F[0]; ii<nn; xx[*ii]=*bb-(sp+sm), ii++, bb++, ai+=n)
    for (jj=in-1, sp=sm=0; ++jj<ii;)
      if((s=ai[*jj]*xx[*jj])<0) sm+=s; else sp+=s;
  for (ii=nn, ai=F[n-1]; !(--ii<in); 
    xx[*ii]=(xx[*ii]-(sp+sm))/ai[*ii], ai-=n)
    for (jj=ii, sp=sm=0; ++jj<nn;)
      if((s=ai[*jj]*xx[*jj])<0) sm+=s; else sp+=s;
  return *this;
}

ls_UINT ls_LU_column::\
post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_LU_column::post_iteration";
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(sizeof(ls_REAL)<8) 
    throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_LU_column::\
residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_LU_column::Lx(const ls_Vector &x) const
{ static char *q1="ls_LU_column::Lx";
  ls_UINT i, j, n=F.number_of_columns(), *in=ind.asArray();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa=F.asArray();
  for(i=0; i<n; vv[i]=xx[i]+s, i++, aa+=n)
    for(j=0, s=0; j<i; s+=xx[j]*aa[*(in+j)], j++);
  return v;
}

ls_Vector ls_LU_column::xL(const ls_Vector &x) const
{ static char *q1="ls_LU_column::xL";
  ls_UINT i, j, n=F.number_of_columns(), *in=ind.asArray(),jj, nn=n-1;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa;
  for(j=0; j<nn; vv[j]=xx[j]+s, j++)
    for(i=j+1, jj=in[j], s=0, aa=F[i]; i<n; 
      s+=xx[i]*aa[jj], i++, aa+=n);
  vv[nn]=xx[nn];
  return v;
}

ls_Vector ls_LU_column::Ux(const ls_Vector &x) const
{ static char *q1="ls_LU_column::Ux";
  ls_UINT i,j,n=F.number_of_columns(),*in=ind.asArray();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa=F.asArray();
  for(i=0; i<n; vv[i]=s, i++, aa+=n)
    for(j=i, s=0; j<n; s+=xx[*(in+j)]*aa[*(in+j)], j++);
  return v;
}

ls_Vector ls_LU_column::xU(const ls_Vector &x) const
{ static char *q1="ls_LU_column::xU";
  ls_UINT i, j, n=F.number_of_columns(), *in=ind.asArray(), jj;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *a=F.asArray(), *aa;
  for(j=0; j<n; vv[jj]=*aa*xx[j]+s, j++)
    for(i=0, jj=in[j], aa=a+jj, s=0; i<j; s+=*aa*xx[i++], aa+=n);
  return v;
}

const ls_LU_column& ls_LU_column::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU_column& ls_LU_column::read_row(istream &f)
{ static char *q1="ls_LU_column::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU_column& ls_LU_column::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU_column& ls_LU_column::read_column(istream &f)
{ static char *q1="ls_LU_column::read_column";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU_column& ls_LU_column::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_LU_column& ls_LU_column::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
