//: $CC -DLS_LIB -c $file
#ifndef LS_LU_DIAGONALC
#define LS_LU_DIAGONALC
#include "ls_names.h"
#include LS_LUDIAGONAL_H

ls_LU_diagonal::ls_LU_diagonal(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); rc=-1;
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  strcpy(F.name,"LU_diagonal-Factorization");
  strcpy(ind.name,"Diagonal-Indices");
}

ls_LU_diagonal::ls_LU_diagonal(ls_LU_diagonal &B)
{ eps=B.eps; strcpy(name=ONAME,B.name); rc=B.rc;
  A_name=A.name; F_name=F.name; ind_name=ind.name;
}

ls_LU_diagonal& ls_LU_diagonal::swap(ls_LU_diagonal &B)
{ char na[ls_len];
  strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); ind.swap(B.ind);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_LU_diagonal& ls_LU_diagonal::\
operator=(const ls_LU_diagonal &B)
{ eps=B.eps; rc=B.rc; ind=B.ind; A=B.A; F=B.F; return *this;}

ls_LU_diagonal::\
ls_LU_diagonal(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_LU_diagonal::ls_LU_diagonal";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  ls_UINT n=A.number_of_columns(), n1=n-1;
  ls_array1<ls_UINT> kind(n,"Diagonal-Indices");ind.swap(kind);
  F=A; eps=ls_EPS;
  if(alpha>0)
  { for(ls_UINT i=0; i<n; F[i][i]+=alpha, i++);
    strcpy(F_name,"regularized_LU_diagonal-Factorization");
  }
  else strcpy(F_name,"LU_diagonal-Factorization");
  ls_UINT i, m=A.number_of_rows(), dim=A.dimension(),
       *in=ind.asArray(), *ii, *jj, *kk, *js, *nn=in+n;
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  double piv, s, e;
  ls_REAL *ai, *ak;
  for (i=0; i<n; in[i]=i, i++);
  for (ii=in, i=0; i<n1; i++, ii++)
  { for(piv=fabs(F[*ii][*ii]), js=jj=ii; ++jj<nn;)
      if((s=fabs(double(F[*jj][*jj])))>piv) piv=s, js=jj;
    if(piv<eps)
    { rc=1+(ls_UINT)(ii-in); F.remove(); ind.remove(); return;}
    m=*ii, *ii=*js, *js=m;
    ai=F[*ii];
    for (piv=double(ai[*ii]), e=piv*eps, piv=1/piv, kk=ii; ++kk<nn;)
    { ak=F[*kk];
      if(fabs(s=piv*ak[*ii])<e){ ak[*ii]=0; continue;}
      for(ak[*ii]=(ls_REAL)s, jj=ii; ++jj<nn; 
        ak[*jj]-=ls_REAL(ai[*jj]*s));
    }
  }
  rc=0;
}

const ls_LU_diagonal& ls_LU_diagonal::\
solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_LU_diagonal::solve";
  ls_UINT *in=ind.asArray(), n=F.number_of_columns(), 
          *ii, *jj, *nn=in+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_REAL s, sp, sm, *xx=x.asArray(), *ai, *bb=b.asArray();
  for (ii=in; ii<nn; xx[*ii]=bb[*ii]-(sp+sm), ii++)
    for (jj=in, sp=sm=0, ai=F[*ii]; jj<ii; jj++)
      if((s=ai[*jj]*xx[*jj])<0) sm+=s; else sp+=s;
  for (ii=nn; !(--ii<in); x[*ii]=(x[*ii]-(sp+sm))/ai[*ii])
    for (jj=ii+1, sp=sm=0, ai=F[*ii]; jj<nn; jj++)
      if((s=ai[*jj]*xx[*jj])<0) sm+=s; else sp+=s;
  return *this;
}

ls_UINT ls_LU_diagonal::\
  post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_LU_diagonal::post_iteration";
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(sizeof(ls_REAL)<8) 
    throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_LU_diagonal::\
residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_LU_diagonal::Lx(const ls_Vector &x) const
{ static char *q1="ls_LU_diagonal::Lx";
  ls_UINT *ii, *jj, n=F.number_of_columns(), 
          *in=ind.asArray(), *nn=in+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa;
  for(ii=in; ii<nn; vv[*ii]=xx[*ii]+s, ii++)
    for(jj=in, s=0, aa=F[*ii]; jj<ii; s+=xx[*jj]*aa[*jj], jj++);
  return v;
}

ls_Vector ls_LU_diagonal::xL(const ls_Vector &x) const
{ static char *q1="ls_LU_diagonal::xL";
  ls_UINT *ii, *jj, n=F.number_of_columns(), 
          *in=ind.asArray(), *nn=in+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray();
  for(jj=in; jj<nn; vv[*jj]=xx[*jj]+s, jj++)
    for(ii=jj+1, s=0; ii<nn; s+=xx[*ii]*F[*ii][*jj], ii++);
  return v;
}

ls_Vector ls_LU_diagonal::Ux(const ls_Vector &x) const
{ static char *q1="ls_LU_diagonal::Ux";
  ls_UINT *ii,*jj,n=F.number_of_columns(),
          *in=ind.asArray(), *nn=in+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa=F.asArray();
  for(ii=in; ii<nn; vv[*ii]=s, ii++)
    for(jj=ii, s=0, aa=F[*ii]; jj<nn; s+=xx[*jj]*aa[*jj], jj++);
  return v;
}

ls_Vector ls_LU_diagonal::xU(const ls_Vector &x) const
{ static char *q1="ls_LU_diagonal::xU";
  ls_UINT *ii,*jj,n=F.number_of_columns(),
          *in=ind.asArray(), *nn=in+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray();
  for(jj=in; jj<nn; vv[*jj]=xx[*jj]*F[*jj][*jj]+s, jj++)
    for(ii=in, s=0; ii<jj; s+=xx[*ii]*F[*ii][*jj], ii++);
  return v;
}

const ls_LU_diagonal& ls_LU_diagonal::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f);ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU_diagonal& ls_LU_diagonal::read_row(istream &f)
{ static char *q1="ls_LU_diagonal::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU_diagonal& ls_LU_diagonal::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU_diagonal& ls_LU_diagonal::read_column(istream &f)
{ static char *q1="ls_LU_diagonal::read_column";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU_diagonal& ls_LU_diagonal::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_LU_diagonal& ls_LU_diagonal::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
