//: $CC -DLS_LIB -c $file
#ifndef LS_LU_ROWC
#define LS_LU_ROWC
#include "ls_names.h"
#include LS_LUROW_H

ls_LU_row::ls_LU_row(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); rc=-1;
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  strcpy(F.name,"LU_row-Factorization");
  strcpy(ind.name,"Row-Indices");
}

ls_LU_row::ls_LU_row(ls_LU_row &B)
{ eps=B.eps; strcpy(name=ONAME,B.name); rc=B.rc;
  A_name=A.name; F_name=F.name; ind_name=ind.name;
}

ls_LU_row& ls_LU_row::swap(ls_LU_row &B)
{ char na[ls_len];
  strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); ind.swap(B.ind);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_LU_row& ls_LU_row::operator=(const ls_LU_row &B)
{ eps=B.eps; rc=B.rc; ind=B.ind; A=B.A; F=B.F; return *this;}

ls_LU_row::ls_LU_row(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_LU_row::ls_LU_row";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  ls_UINT i, j, k, n=A.number_of_columns(), m=A.number_of_rows();
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  ls_array1<ls_UINT> kind(n,"Row-Indices"); ind.swap(kind);
  F=A; eps=ls_EPS;
  if(alpha>0)
  { for(ls_UINT i=0; i<n; F[i][i]+=alpha, i++);
    strcpy(F_name,"regularized_LU_row-Factorization");
  }
  else strcpy(F_name,"LU_row-Factorization");
  ls_UINT *in=ind.asArray(), *ii, *jj, *js, *je=in+n, n1=n-1;
  ls_Vector x(n);
  double piv, s, e;
  ls_REAL *xx=x.asArray(), *ai, *ak, *a=F.asArray();
  for (i=0, ai=a; i<n; in[i]=i, xx[i]=ls_REAL(1./s), i++)
    for (s=0, j=0; j<n; s+=fabs(double(*(ai++))), j++);
  for (i=0, ii=in; i<n1; i++, ii++)
  { for(piv=fabs(F[*ii][i]*xx[*ii]), js=jj=ii; ++jj<je;)
      if((s=fabs(F[*jj][i]*xx[*jj]))>piv) piv=s, js=jj;
    if(piv<eps){ rc=1+i; F.remove(); ind.remove(); return;}
    m=*ii, *ii=*js, *js=m;
    ai=F[*ii];
    for(e=ai[i]*eps, piv=1./ai[i], k=i; ++k<n;)
    { ak=F[in[k]];
      if(fabs(s=ak[i]*piv)<e){ ak[i]=0; continue;}
      for (ak[i]=(ls_REAL)s, j=i; ++j<n; ak[j]-=ls_REAL(ai[j]*s));
    }
  }
  rc=0;
}

const ls_LU_row& ls_LU_row::\
solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_LU_row::solve";
  ls_UINT i, j, *inr=ind.asArray(), *ir, n=F.number_of_rows() ;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_REAL s, sp, sm, *ai, *xx=x.asArray(), *bb=b.asArray();
  for(i=0, ir=inr; i<n; xx[i]=bb[*ir]-(sp+sm), i++, ir++)
    for(ai=F[*ir], j=0, sp=sm=0; j<i; j++)
      if((s=ai[j]*xx[j])<0) sm+=s; else sp+=s;
  for (i=n, ir=inr+n-1; i--; xx[i]=(xx[i]-(sp+sm))/ai[i], ir--)
    for (ai=F[*ir], j=i, sp=sm=0; ++j<n;)
      if((s=ai[j]*xx[j])<0) sm+=s; else sp+=s;
  return *this;
}

ls_UINT ls_LU_row::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_LU_row::post_iteration";
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(sizeof(ls_REAL)<8) 
    throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_LU_row::\
residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_LU_row::Lx(const ls_Vector &x) const
{ static char *q1="ls_LU_row::Lx";
  ls_UINT i, j, *ir, n=F.number_of_columns(), 
          *inr=ind.asArray(), *nr=inr+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *aa, *xx=x.asArray(), *vv=v.asArray();
  for(i=0, ir=inr; i<n; vv[*ir]=xx[i]+s, i++, ir++)
    for(j=0, s=0, aa=F[*ir]; j<i; s+=*(aa++)*xx[j], j++);
  return v;
}

ls_Vector ls_LU_row::xL(const ls_Vector &x) const
{ static char *q1="ls_LU_row::xL";
  ls_UINT j, *ir, *ii, n=F.number_of_columns(), 
          *inr=ind.asArray(), *nr=inr+n;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray();
  for(j=0, ir=inr; j<n; vv[j]=xx[*ir]+s, j++, ir++)
    for(ii=ir, s=0; ++ii<nr; s+=xx[*ii]*F[*ii][j]);
  return v;
}

ls_Vector ls_LU_row::Ux(const ls_Vector &x) const
{ static char *q1="ls_LU_row::Ux";
  ls_UINT i, j, n=F.number_of_columns(), *in=ind.asArray();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa;
  for(i=0; i<n; vv[i]=s, i++)
    for(j=i, s=0, aa=&F[*(in+i)][j]; j<n; s+=*(aa++)*xx[j++]);
  return v;
}

ls_Vector ls_LU_row::xU(const ls_Vector &x) const
{ static char *q1="ls_LU_row::xU";
  ls_UINT i, j, *ii, n=F.number_of_columns(), *in=ind.asArray();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray();
  for(j=0; j<n; vv[j]=xx[j]*F[*ii][j]+s, j++)
    for(i=0, ii=in, s=0; i<j; s+=xx[i]*F[*ii][j], i++, ii++);
  return v;
}

const ls_LU_row& ls_LU_row::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU_row& ls_LU_row::read_row(istream &f)
{ static char *q1="ls_LU_row::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU_row& ls_LU_row::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); ind.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU_row& ls_LU_row::read_column(istream &f)
{ static char *q1="ls_LU_column::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU_row& ls_LU_row::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_LU_row& ls_LU_row::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
