//: $CC -DLS_LIB -c $file
#ifndef LS_LU_TOTALC
#define LS_LU_TOTALC
#include "ls_names.h"
#include LS_LUTOTAL_H

ls_LU_total::ls_LU_total(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); rc=-1;
  A_name=A.name; F_name=F.name; indr_name=indr.name; 
  indc_name=indc.name;
  strcpy(F.name,"LU_total-Factorization");
  strcpy(indr.name,"Row-Indices"); 
  strcpy(indc.name,"Column-Indices");
}

ls_LU_total::ls_LU_total(ls_LU_total &B)
{ eps=B.eps; strcpy(name=ONAME,B.name); rc=B.rc;
  A_name=A.name; F_name=F.name; indr_name=indr.name; 
  indc_name=indc.name;
}

ls_LU_total& ls_LU_total::swap(ls_LU_total &B)
{ char na[ls_len];strcpy(na,name);strcpy(name,B.name);
  strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); indr.swap(B.indr); 
  indc.swap(B.indc);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_LU_total& ls_LU_total::operator=(const ls_LU_total &B)
{ eps=B.eps; rc=B.rc; indr=B.indr; indc=B.indc; A=B.A; F=B.F;
  return *this;
}

ls_LU_total::ls_LU_total(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_LU_total::ls_LU_total";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; indr_name=indr.name; 
  indc_name=indc.name;
  ls_UINT n=A.number_of_columns(), n1=n-1;
  ls_array1<ls_UINT> kindr(n,"Row-Indices");indr.swap(kindr);
  ls_array1<ls_UINT> kindc(n,"Column-Indices");indc.swap(kindc);
  F=A; eps=ls_EPS;
  if(alpha>0)
  { for(ls_UINT i=0; i<n; F[i][i]+=alpha, i++);
    strcpy(F_name,"regularized_LU_total-Factorization");
  }
  else strcpy(F_name,"LU_total-Factorization");
  ls_Vector rrow(n), ccol(n);
  ls_UINT i,j,m=A.number_of_rows(), dim=A.dimension(),
       *inr=indr.asArray(),*inc=indc.asArray(),
       *jj,*kk, *ip, *jp, *ir, *jc, *nc=inc+n,
       *ni=inr+n-1, *nk=inr+n;
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  double piv, s, e;
  ls_REAL *ai, *ak, *row=rrow.asArray(), *col=ccol.asArray();
  for (i=0; i<n; inr[i]=inc[i]=i, row[i++]=s)
    for(j=0, ai=F[i],s=0; j<n; j++)
      s+=fabs(double(ai[j])), col[j]+=(ls_REAL)fabs(ai[j]);
  for (ir=inr, jc=inc; ir<ni; ir++, jc++)
  { for(piv=0, ip=ir, jp=jc, kk=ir-1; ++kk<nk;)
      for(jj=jc-1, ak=F[*kk], e=row[*kk]; ++jj<nc;)
      if((s=fabs(double(ak[*jj]))/(e*col[*jj]))>piv)
        piv=s, ip=kk, jp=jj;
    if(piv<eps)
    { rc=1+(ls_UINT)(ir-inr); F.remove(); indr.remove(); 
      indc.remove(); return;
    }
    m=*ir, *ir=*ip, *ip=m; m=*jc, *jc=*jp, *jp=m;
    ai=F[*ir]; piv=ai[*jc];
    for (e=piv*eps, piv=1/piv, kk=ir; ++kk<nk;)
    { ak=F[*kk]; jj=jc;
      if(fabs(s=piv*ak[*jj])<e){ ak[*jj]=0; continue;}
      for(ak[*jj]=s; ++jj<nc; ak[*jj]-=ai[*jj]*s);
    }
  }
  rc=0;
}

const ls_LU_total& ls_LU_total::\
solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_LU_total::solve";
  ls_UINT *inr=indr.asArray(),*inc=indc.asArray(),
          n=F.number_of_columns(),
       *ir, *jc, *nr=inr+n, *nc=inc+n, *je;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_REAL s, sp, sm, *xx=x.asArray(), *ai, *bb=b.asArray();
  for (ir=inr-1, je=inc; ++ir<nr; xx[*jc]=bb[*ir]-(sp+sm), je++)
    for (jc=inc-1, sp=sm=0, ai=F[*ir]; ++jc<je;)
      if((s=ai[*jc]*xx[*jc])<0) sm+=s; else sp+=s;
  for (ir=nr, je=nc-1; !((--ir)<inr); 
    xx[*je]=(xx[*je]-(sp+sm))/ai[*je], je--)
    for (jc=je, sp=sm=0, ai=F[*ir]; ++jc<nc;)
      if((s=ai[*jc]*xx[*jc])<0) sm+=s; else sp+=s;
  return *this;
}

ls_UINT ls_LU_total::\
  post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_LU_total::post_iteration";
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(sizeof(ls_REAL)<8) 
    throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(),*ha=h.asArray(),*xa=x.asArray(),
          *vv,*hh,*xx,*xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_LU_total::\
residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_LU_total::Lx(const ls_Vector &x) const
{ static char *q1="ls_LU_total::Lx";
  ls_UINT *ir, *jc, *nr, *nc, *inc, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa;
  for(ir=indr.asArray(), inc=indc.asArray(), nr=ir+n, nc=inc;
    ir<nr; vv[*ir]=xx[*jc]+s, ir++, nc++)
    for(jc=inc, s=0, aa=F[*ir]; jc<nc; s+=xx[*jc]*aa[*jc], jc++);
  return v;
}

ls_Vector ls_LU_total::xL(const ls_Vector &x) const
{ static char *q1="ls_LU_total::xL";
  ls_UINT *ir, *jc, *nr, *nc, *inr, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray();
  for(jc=indc.asArray(), inr=indr.asArray(), nc=jc+n, nr=inr+n;
      jc<nc; vv[*jc]=xx[*ir]+s, inr++, jc++)
    for(ir=inr, s=0; ++ir<nr; s+=xx[*ir]*F[*ir][*jc]);
  return v;
}

ls_Vector ls_LU_total::Ux(const ls_Vector &x) const
{ static char *q1="ls_LU_total::Ux";
  ls_UINT *ir,*jc, *nc, *nr, *inc, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray(), *aa=F.asArray();
  for(ir=indr.asArray(), nr=ir+n, inc=indc.asArray(), nc=inc+n;
      ir<nr; vv[*ir]=s, ir++, inc++)
    for(jc=inc-1, s=0, aa=F[*ir]; ++jc<nc; s+=xx[*jc]*aa[*jc]);
  return v;
}

ls_Vector ls_LU_total::xU(const ls_Vector &x) const
{ static char *q1="ls_LU_total::xU";
  ls_UINT n=F.number_of_columns(), *inr, *ir, *jc, *nr, *nc;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n);
  ls_REAL s, *xx=x.asArray(), *vv=v.asArray();
  for(jc=indc.asArray(), inr=indr.asArray(), nc=jc+n, nr=inr;
      jc<nc; vv[*jc]=xx[*ir]*F[*ir][*jc]+s, jc++, nr++)
    for(ir=inr-1, s=0; ++ir<nr; s+=xx[*ir]*F[*ir][*jc]);
  return v;
}

const ls_LU_total& ls_LU_total::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f);indr.write(f);indc.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU_total& ls_LU_total::read_row(istream &f)
{ static char *q1="ls_LU_total::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(indr.name)=0, 
       *(indc.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); indr.read(f); indc.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU_total& ls_LU_total::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); indr.write(f); indc.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_LU_total& ls_LU_total::read_column(istream &f)
{ static char *q1="ls_LU_total::read_column";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, 
       *(indr.name)=0, *(indc.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); indr.read(f); indc.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_LU_total& ls_LU_total::operator>> (char *nam) const
{ if(!*nam) write_row(cout);
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.4)), write_row(f);
  }
  return *this;
}

ls_LU_total& ls_LU_total::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
