//: $CC -DLS_LIB -c $file
#ifndef LS_MATRIXC
#define LS_MATRIXC
#include "ls_names.h"
#include LS_MATRIX_H

ls_Matrix::ls_Matrix(ls_UINT mm, ls_UINT nn, char *nam):
ls_array2<ls_REAL>(mm, nn, nam){}

ls_Matrix::ls_Matrix(ls_Matrix &A):ls_array2<ls_REAL>(A){}

const ls_Matrix& ls_Matrix::operator= (const ls_Matrix &B)
{ ls_array2<ls_REAL>::operator=(B); return *this;}

ls_Matrix& ls_Matrix::operator+=(const ls_Matrix &B)
{ ls_REAL *ba=B.asArray(), *be=ba+dim, *aa=a;
  while(ba<be) *aa+=*ba, aa++, ba++;
  return *this;
}

ls_Matrix ls_Matrix::operator-(const ls_Matrix &B) const
{ ls_Matrix C(m,n); C=*this; C-=B; return C;}

ls_Matrix& ls_Matrix::operator-=(const ls_Matrix &B)
{ ls_REAL *ba=B.asArray(), *be=ba+dim, *aa=a;
  while(ba<be) *aa-=*ba, aa++, ba++;
  return *this;
}

ls_Matrix ls_Matrix::operator+ (const ls_Matrix &B) const
{ ls_Matrix C(m,n); C=*this; C+=B; return C;}

ls_Matrix ls_Matrix::operator*(const ls_Matrix &B) const
{ ls_Matrix C(m, B.number_of_columns());
  for(ls_UINT i=0; i<m; C.put_row(row(i)*B,i), i++);
  return C;
}

ls_Matrix ls_Matrix::operator*(ls_REAL s) const
{ ls_Matrix B(m,n); B=*this;
  ls_REAL *bb=B.asArray(), *be=bb+dim;
  while(bb<be) *bb*=s, bb++;
  return B;
}


ls_Vector ls_Matrix::operator*(const ls_Vector &x) const
{ ls_Vector y(m); ls_array1<double> u(n);
  ls_REAL *aa=a, *yy=y.asArray(), *ye=yy+m, *xx=x.asArray();
  double s, sm, sp, *ua=u.asArray(), *uu, *ue=ua+n;
  uu=ua; while(uu<ue) *uu=*xx, xx++, uu++;
  for (; yy<ye; *yy=(ls_REAL)(sp+sm), yy++)
    for (sp=sm=0, uu=ua; uu<ue; aa++, uu++)
      if((s=*uu*(*aa))<0) sm+=s; else sp+=s;
  return y;
}

ls_Matrix& ls_Matrix::operator*=(ls_REAL s)
{ for(ls_REAL *aa=a, *ae=aa+dim; aa<ae; *aa*=s, aa++); return *this;}

ls_UINT ls_Matrix::solve(ls_Vector &x, ls_Vector &b)const
{ b-=(*this)*x; ls_Vector p; p=b*(*this);
  ls_REAL s, bb=b*b, bbn, pp=p*p, eps=x.eps;
  ls_UINT it=0, rep=n*n;
  while (!((bb+pp)<eps)&&(it++<rep))
  { s=bb/pp, x+=s*p;
    b-=(s*(*this)*p);
    bbn=b*b;
    p=b*(*this)+(bbn/bb)*p,  pp=p*p, bb=bbn;
  }
  return it;
}

ls_Matrix operator*(ls_REAL x, const ls_Matrix &B)
{ ls_UINT m=B.number_of_rows(), n=B.number_of_columns(), 
          dim=B.dimension();
  ls_Matrix C(m,n);
  ls_REAL *cc=C.asArray(), *bb=B.asArray(), *be=bb+dim;
  while(bb<be) *cc=*bb*x, bb++, cc++;
  return C;
}

ls_Vector operator*(const ls_Vector &x, const ls_Matrix &B)
{ ls_UINT m=B.number_of_rows(), n=B.number_of_columns();
  ls_array1<double> u(n); ls_Vector y(n);
  ls_REAL *A=B.asArray(), *aa, *yy=y.asArray(), *ye=yy+n, 
          *xx=x.asArray();
  double s, sm, sp, *ua=u.asArray(), *uu, *ue=ua+n;
  uu=ua; while(uu<ue) *uu=*xx, xx++, uu++;
  for (; yy<ye; *yy=(ls_REAL)(sp+sm), A++, yy++)
    for (aa=A, sp=sm=0, uu=ua; uu<ue; aa+=n, uu++)
      if((s=*(uu)*(*aa))<0) sm+=s; else sp+=s;
  return y;
}

ls_Matrix ls_Vector::dyad(const ls_Vector &v) const
{ ls_UINT vdim=v.dimension();
  ls_Matrix A(dim,vdim);
  ls_REAL *va=v.asArray(), *vv, *uu, *ve=va+vdim, *ue=a+dim, 
          *aa=A.asArray();
  for(uu=a; uu<ue; uu++) 
    for (vv=va; vv<ve; *aa=*uu*(*vv), aa++, vv++);
  return A;
}
#endif
