//: $CC -DLS_LIB -c $file
#ifndef LS_QRC
#define LS_QRC
#include "ls_names.h"
#include LS_QR_H

ls_QR::ls_QR(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); rc=-1;
  A_name=A.name; F_name=F.name; gamma_name=gamma.name; 
  rho_name=rho.name;
  strcpy(F.name,"QR-Factorization");
  strcpy(gamma.name,"gamma-array");
  strcpy(rho.name,"rho-array");
}

ls_QR::ls_QR(ls_QR &B)
{ eps=B.eps; strcpy(name=ONAME,B.name); rc=B.rc;
  A_name=A.name; F_name=F.name; gamma_name=gamma.name; 
  rho_name=rho.name;
}

ls_QR& ls_QR::swap(ls_QR &B)
{ char na[ls_len];
  strcpy(na,name);strcpy(name,B.name);strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); gamma.swap(B.gamma);
  rho.swap(B.rho);
  ls_UINT r=rc; rc=B.rc; B.rc=r;
  return *this;
}

const ls_QR& ls_QR::operator=(const ls_QR &B)
{ gamma=B.gamma; rc=B.rc; rho=B.rho; eps=B.eps; A=B.A; F=B.F;
  return *this;
}

ls_QR::ls_QR(ls_Matrix &B, ls_REAL alpha, char *nam)
{ static char *q1="ls_QR::ls_QR";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; gamma_name=gamma.name; 
  rho_name=rho.name;
  ls_UINT i, j, nn, dim=A.dimension(), n=A.number_of_columns(),
       m=A.number_of_rows();
  if(m<n) throw ls_error(q1,name,size_error_);
  ls_array1<ls_REAL> kgamma(n,"gamma-array"), krho(n,"rho-array");
  gamma.swap(kgamma); rho.swap(krho);
  F=A; eps=ls_EPS;
  if(alpha>0)
  {  for (i=0; i<n; F[i][i]+=alpha, i++);
     strcpy(F_name,"regularized_QR-Factorization");
  }
  else strcpy(F_name,"QR-Factorization");
  double s, sm, sp;
  ls_REAL *aa=F.asArray(), *bb, *ae=aa+dim, *g=gamma.asArray(), 
          *r=rho.asArray();
  nn=(m>n)? n : n-1;

  for(i=0; i<nn; i++, aa+=n, g++, r++)
  { for (s=0, bb=aa+i; bb<ae; s+=*bb*(*bb), bb+=n);
    if((s=sqrt(s))<eps)
    { rc=1+i; F.remove(); rho.remove(); gamma.remove(); return;}
    if(aa[i]>0) s=-s;
    for(*r=(ls_REAL)s, aa[i]-=(ls_REAL)s, 
        *g=(ls_REAL)(-1./(s*aa[i])), j=i+1; j<n; j++)
    { for(sm=sp=0, bb=aa; bb<ae; bb+=n)
        if((s=bb[j]*bb[i])<0) sm+=s; else sp+=s;
      for(s=*g*(sp+sm), bb=aa; bb<ae; 
        bb[j]-=ls_REAL(s*bb[i]), bb+=n);
    }
  }
  if(nn<n) *r=aa[nn];
  rc=0;
}

const ls_QR& ls_QR::solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="ls_QR::solve";
  ls_UINT i, k, nn=b.dimension(), n=F.number_of_columns(), 
          m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(nn!=m) throw ls_error(q1,b.name,size_error_);
  ls_Vector z; z=b;
  ls_REAL s,sp,sm,*a=F.asArray(),*aa,*bb,*za=z.asArray(),
          *zz,*ze=za+m,
          *g=gamma.asArray(), *xx=x.asArray(), *r=rho.asArray();
  nn=(m>n)? n : n-1;
  for (k=0, aa=a; k<nn; k++, aa+=n, g++)
  { for (sp=sm=0, zz=za+k, bb=aa; zz<ze; zz++, bb+=n)
      if((s=bb[k]*(*zz))<0) sm+=s; else sp+=s;
    for (s=*g*(sp+sm), bb=aa, zz=za+k; zz<ze; 
         *zz-=s*bb[k], zz++, bb+=n);
  }
  if(nn!=n) nn++;
  for (k=nn, aa=F[k-1]; k--; xx[k]=(za[k]-(sp+sm))/r[k], aa-=n)
    for (i=k+1, sp=sm=0; i<n; i++)
      if((s=aa[i]*xx[i])<0) sm+=s; else sp+=s;
  return *this;
}

ls_UINT ls_QR::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="ls_QR::post_iteration";
  ls_UINT rep=10, l=0, n=F.number_of_columns();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(sizeof(ls_REAL)<8) 
    throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_Vector v(n), h(n);
  ls_REAL *va=v.asArray(), *ha=h.asArray(), *xa=x.asArray(),
          *vv,*hh,*xx, *xe=xa+n;
  for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
    if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));

  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while(!(r*r<eps)&&(l++<rep))
  { solve(z,r); x+=z;
    for(xx=xa,hh=ha,vv=va; xx<xe; xx++,xx++,vv++)
      if(fabs(*xx)<1.) *hh=*xx, *vv=0; else *hh=*xx-(*vv=float(*xx));
    r=(b-A*v)-A*h;
  }
  b=r; return l;
}

ls_Vector ls_QR::\
residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

ls_Vector ls_QR::Qx(const ls_Vector &x) const
{ static char *q1="ls_QR::Qx";
  ls_UINT i, j, n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  ls_Vector v; v=x;
  ls_REAL s,*vv=v.asArray(), *gg=gamma.asArray(), *aa;
  for(j=n-1; j--;)
  { for (i=j, s=0, aa=&F[i][i]; i<n; s+=*aa*vv[i], i++, aa+=n);
    for (i=j, s*=gg[i], aa=&F[i][i]; i<n; vv[i]-=*aa*s, i++, aa+=n);
  }
  return v;
}

ls_Vector ls_QR::xQ(const ls_Vector &x) const
{ static char *q1="ls_QR::xQ";
  ls_UINT i, j, n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  ls_Vector v; v=x;
  ls_REAL s,*vv=v.asArray(), *gg=gamma.asArray(), *aa;
  for(j=0; j<n-1; j++)
  { for (s=0, i=j, aa=&F[i][j]; i<n; s+=*aa*vv[i], i++, aa+=n);
    for (i=j, s*=gg[i], aa=&F[i][i]; i<n; vv[i]-=*aa*s, i++, aa+=n);
  }
  return v;
}

ls_Vector ls_QR::Rx(const ls_Vector &x) const
{ static char *q1="ls_QR::Rx";
  ls_UINT i,j,n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  ls_Vector v(n);
  ls_REAL s,*xx=x.asArray(),*vv=v.asArray(),*aa,*r=rho.asArray();
  for(i=0; i<n; vv[i]=r[i]*xx[i]+s, i++)
    for(j=i+1, aa=&F[i][i], s=0; j<n; s+=*(aa++)*xx[j], j++);
  return v;
}

ls_Vector ls_QR::xR(const ls_Vector &x) const
{ static char *q1="ls_QR::xR";
  ls_UINT i,j,n=F.number_of_columns(), m=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  if(m!=n) throw ls_error(name,q1,object_not_quadratic_);
  ls_Vector v(n);
  ls_REAL s,*xx=x.asArray(),*vv=v.asArray(),*r=rho.asArray(), *aa;
  for(j=0; j<n; vv[j]=xx[j]*r[j]+s, j++)
    for(i=0, s=0, aa=&F[i][j]; i<j; s+=*aa*xx[i], i++, aa+=n);
  return v;
}

const ls_QR& ls_QR::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f); F.write_row(f); gamma.write(f); rho.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_QR& ls_QR::read_row(istream &f)
{ static char *q1="ls_QR::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(rho.name)=0, 
       *(gamma.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f); F.read_row(f); gamma.read(f); rho.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_QR& ls_QR::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); gamma.write(f); rho.write(f);
  f<<"}"<<endl;
  return *this;
}

ls_QR& ls_QR::read_column(istream &f)
{ static char *q1="ls_LU_column::read_row";
  char w[ls_len];
  if(*name) while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(rho.name)=0, 
       *(gamma.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f); F.read_column(f); gamma.read(f); rho.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const ls_QR& ls_QR::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

ls_QR& ls_QR::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
