//: $CC -DLS_LIB -c $file
#ifndef LS_ARRAY1MC
#define LS_ARRAY1MC
#include "ls_names.h"
#include LS_ARRAY1M_H

template<class T>
ls_array1M<T>::ls_array1M(ls_UINT nn, char *nam)
{ strcpy(name=ONAME,nam); if(!(n=nn)) return;
  a=new array1M<T>[n], memset(a, 0, sizeof(array1M<T>)*n);
}

template<class T>
ls_array1M<T>::ls_array1M(ls_array1M<T> &v)
{ strcpy(name=ONAME,v.name);
  n=v.number_of_arrays(); a=v.asArray(); v.set_data();
}

template<class T>
ls_array1M<T>::~ls_array1M()
{ if(!n) return;
  array1M<T> *aa=a, *ae=aa+n;
  while(aa<ae){ if(aa->l) delete[]aa->f; aa++;}
  delete[]a;
}

template<class T>
ls_array1M<T>& ls_array1M<T>::swap(ls_array1M<T> &v)
{ char na[ls_len];
  strcpy(na,name);strcpy(name,v.name);strcpy(v.name,na);
  ls_UINT vn=v.number_of_arrays(); array1M<T> *aa=v.asArray();
  v.set_data(a,n); n=vn; a=aa; return *this;
}

template<class T>
const ls_array1M<T>& ls_array1M<T>::operator=(const ls_array1M<T> &v)
{ static char *q1="ls_array1M::=";
  array1M<T> *va=v.asArray(), *aa=a, *ae=aa+n;
  ls_UINT vn=v.number_of_arrays(), l;
  if(!vn) throw ls_error(v.name,q1,object_is_empty_);
  if(a==va) return *this;
  if(!n) a=new array1M<T>[n=vn], memset(a,0,sizeof(array1M<T>)*n);
  if(n!=vn) throw ls_error(name,q1,size_error_);
  while (aa<ae)
  { if((l=va->l)!=aa->l)
    { if(aa->l) delete []aa->f;
      if(l) aa->f=new T[l];
      aa->l=l;
    }
    if(l) memcpy(aa->f, va->f, sizeof(T)*l);
    aa++, va++;
  }
  return *this;
}

template<class T>
ls_array1M<T>& ls_array1M<T>::put(T v, ls_UINT i)
{ static char *q1="ls_array1M::put";
  if(!(i<n)) ls_error(name,q1,index_error_);
  array1M<T> *aa=a+i;
  if(aa->l!=1)
  { if(aa->l) delete[]a->f;
    aa->f=new T[aa->l=1];
  }
  *(aa->f)=v;
  return *this;
}

template<class T>
ls_array1M<T>& ls_array1M<T>::put(const T *v, ls_UINT l, ls_UINT i)
{ static char *q1="ls_array1M::put";
  if(!(i<n)) ls_error(name,q1,index_error_);
  array1M<T> *aa=a+i;
  if(aa->l!=l)
  { if(aa->l) delete[]aa->f;
    aa->f=new T[aa->l=l];
  }
  memcpy(aa->f, v, sizeof(T)*l);
}

template<class T>
ls_array1M<T>& ls_array1M<T>::\
put_array(const ls_array1<T> &v, ls_UINT i)
{ static char *q1="ls_array1M::put_array";
  if(!(i<n)) ls_error(name,q1,index_error_);
  ls_UINT vn=v.dimension();
  if(!vn) throw ls_error(v.name,q1,object_is_empty_);
  if(vn>n) vn=n;
  array1M<T> *aa=a+i;
  if(aa->l!=vn)
  { if(aa->l) delete[]aa->f;
    aa->f=new T[vn], aa->l=vn;
  }
  memcpy(aa->f, v.asArray(), sizeof(T)*vn);
  return *this;
}

template<class T>
const ls_array1M<T>& ls_array1M<T>::\
get_array(ls_array1<T> &v, ls_UINT i) const
{ static char *q1="ls_array1M::get_array";
  if(!(i<n)) ls_error(name,q1,index_error_);
  ls_UINT vn=v.dimension();
  if(!vn) throw ls_error(nam,q1,object_is_empty_);
  if(vn>(a+i)->i) vn=(a+i)->i;
  memcpy(v.asArray(), (a+i)->f, sizeof(T)*vn);
  return *this;
}

template<class T>
ls_array1M<T>& ls_array1M<T>::remove_array(ls_UINT i)
{ static char *q1="ls_array1M::remove";
  if(!n) throw ls_error(name,q1,object_is_empty_);
  if(!(i<n)) throw ls_error(name,q1,index_error_);
  if((a+i)->l) delete[]aa->f, (a+i)->l=0;
  return *this;
}

template<class T>
ls_array1M<T>& ls_array1M<T>::remove()
{ static char *q1="ls_array1M::remove";
  if(!n) return *this;
  array1M<T> *aa=a, *ae=aa+n;
  while(aa<ae){ if(aa->l) delete[]aa->f; aa++;}
  delete[]a; n=0;
  return *this;
}

template<class T>
const ls_array1M<T>& ls_array1M<T>::write(ostream &f) const
{ static char *q1="ls_array1M::write";
  if(!n) throw ls_error(q1,name,object_is_empty_);
  ls_UINT i=0; T *c, *ce;
  f <<name<<endl<<"{ "<<number_of_arrays_<<" "<<n<<endl;
  for (array1M<T> *aa=a, *ae=a+n; aa<ae; aa++, i++)
  { f<<"array: "<<i<<" "<<number_of_elements_<<" "<<aa->l<<endl;
      for (c=aa->f, ce=c+aa->l; c<ce; c++)
        if(*c<0) f << " " << *c; else f << "  " << *c;
    if(aa->l) f<<endl;
  }
  f<<"}"<<endl;
  return *this;
}

template<class T>
ls_array1M<T>& ls_array1M<T>::read(istream &f)
{ static char *q1="ls_array1M::read";
  if(n) throw ls_error(name, q1, object_not_empty_);
  char w[ls_len]; ls_UINT i, l;
  array1M<T> *aa;
  T *c;
  if(*name) while(f>>w&&strcmp(w,name)); else f>>name;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  if(!(f>>w)||strcmp(w, number_of_arrays_)||!(f>>n))
    throw ls_error(name,q1,object_contents_error_);
  a=new array1M<T>[n],memset(a,0,sizeof(array1M<T>)*n);
  while ((f>>w)&&(*w!='}'))
  { if(strcmp(w, "array:")||!(f>>i)||(i>n)\
       ||!(f>>w)||strcmp(w, number_of_elements_)||!(f>>l))
      throw ls_error(name,q1,object_contents_error_);
    aa=a+i;
    if(!l) continue;
    if(!a->l) aa->f=new T[l], aa->l=l;
    if(l!=aa->l) throw ls_error(name,q1,size_error_);
    c=aa->f; while (l--) f>>*c, c++;
  }
  if(*w!='}') throw ls_error(name,q1,object_end_error_);
  return *this;
}

template<class T>
const ls_array1M<T>& ls_array1M<T>::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

template<class T>
ls_array1M<T>& ls_array1M<T>::operator<< (char *nam)
{ if(!*nam) read(cin); else{ ifstream f(nam); read(f);}
  return *this;
}
#endif
