//: $CC -DLS_LIB -c $file
#ifndef LS_ARRAYUC
#define LS_ARRAYUC
#include "ls_names.h"
#include LS_ARRAYU_H

template<class T>
ls_arrayU<T>::ls_arrayU(ls_UINT nn,char *nam):
n(nn),ls_array1<T>((nn*(nn+1))/2,nam){}

template<class T>
ls_arrayU<T>::ls_arrayU(ls_arrayU<T> &A):
ls_array1<T>(A),n(A.dimension()){ A.set_data();}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::swap(ls_arrayU<T> &A)
{ ls_array1<T>::swap(A); ls_UINT nn=n; n=A.dimension();
  A.set_data(nn); return *this;
}

template<class T>
T* ls_arrayU<T>::operator[](ls_UINT k) const
{
#ifdef LS_TEST
  static char *q1="ls_arrayU::[]";
  if(!(k<n)) throw ls_error(name,q1,index_error_);
#endif
  ls_UINT kk=n-k; return a+dim-(kk*(kk+1))/2-k;
}

template<class T>
const ls_arrayU<T>& ls_arrayU<T>::operator=(const ls_arrayU<T> &v)
{ static char *q1="ls_arrayU::=";
  ls_UINT vdim=v.dimension();
  if(!vdim) throw ls_error(v.name,q1,object_is_empty_);
  if(!n) n=vdim;
  if(n!=vdim) throw ls_error(name,q1, size_error_);
  ls_array1<ls_REAL>::operator=(v); return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::put(T v, ls_UINT i, ls_UINT j)
{
#ifdef LS_TEST
  static char *q1="ls_arrayU::put";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
#endif
  (*this)[i][j]=v; return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::put_row(const ls_array1<T> &v, 
                                    ls_UINT i, ls_UINT j)
{ static char *q1="ls_arrayU::put_row";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  ls_UINT vdim=v.dimension(), ll=(n-j<vdim)? n-j:vdim;
  for(T *vv=v.asArray(),*ve=vv+ll,*aa=(*this)[i]+j;
      vv<ve;*aa=*vv,aa++,vv++);
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::put_row(T s, ls_UINT i, ls_UINT j)
{ static char *q1="ls_arrayU::put_row";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  for(T *aa=(*this)[i]+j,*ae=aa+n-j; aa<ae; *aa=s, aa++);
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::put_column(const ls_array1<T> &v, 
                                       ls_UINT i, ls_UINT j)
{ static char *q1="ls_arrayU::put_column";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  ls_UINT vdim=v.dimension(), ll=(j-i+1<vdim)? j-i+1:vdim;
  T *vv=v.asArray();
  while(ll--) (*this)[i++][j]=*vv, vv++;
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::put_column(T s, ls_UINT i, ls_UINT j)
{ static char *q1="ls_arrayU::put_column";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  for(; i<=j;(*this)[i++][j]=s);
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::put_diagonal(const ls_array1<T> &v, 
                                         ls_UINT i,ls_UINT j)
{ static char *q1="ls_arrayU::put_diagonal";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  ls_UINT vdim=v.dimension(), ll=(n-j-i<vdim)? n-j-i:vdim;
  T *vv=v.asArray();
  while(ll--) (*this)[i++][j++]=*vv, vv++;
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::put_diagonal(T s, ls_UINT i,ls_UINT j)
{ static char *q1="ls_arrayU::put_diagonal";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  for(; j<n; (*this)[i++][j++]=s);
  return *this;
}

template<class T>
T ls_arrayU<T>::get(ls_UINT i, ls_UINT j) const
{
#ifdef LS_TEST
  static char *q1="ls_arrayU::get";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
#endif
  return (*this)[i][j];
}

template<class T>
const ls_arrayU<T>& ls_arrayU<T>::\
get_row(ls_array1<T> &v, ls_UINT i, ls_UINT j) const
{ static char *q1="ls_arrayU::get_row";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  ls_UINT vdim=v.dimension(), ll=(n-j<vdim)? n-j:vdim;
  T *vv=v.asArray(),*aa=(*this)[i]+j;
  while(ll--) *vv=*aa, aa++, vv++;
  return *this;
}

template<class T>
const ls_arrayU<T>& ls_arrayU<T>::\
get_column(ls_array1<T> &v, ls_UINT i, ls_UINT j) const
{ static char *q1="ls_arrayU::get_column";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  ls_UINT vdim=v.dimension(), ll=(j-i+1<vdim)? j-i+1:vdim;
  T *vv=v.asArray();
  while(ll--) *vv=(*this)[i++][j], vv++;
  return *this;
}

template<class T>
const ls_arrayU<T>& ls_arrayU<T>::\
get_diagonal(ls_array1<T> &v, ls_UINT i,ls_UINT j) const
{ static char *q1="ls_arrayU::get_diagonal";
  if(!(j<n)||!(i<n)||(i>j)) throw ls_error(name,q1,index_error_);
  ls_UINT vdim=v.dimension(), ll=(n-j-i<vdim)? n-j-i:vdim;
  T *vv=v.asArray();
  while(ll--) *vv=(*this)[i++][j++], vv++;
  return *this;
}

template<class T>
const ls_arrayU<T>& ls_arrayU<T>::get_array(ls_array2<T> &A) const\
{ static char *q1="ls_arrayU::get_array";
  if(!n) throw ls_error(name,q1,object_is_empty_);
  if(A.number_of_columns()!=A.number_of_rows()) 
    throw ls_error(name,q1,object_not_quadratic_);
  if(n!=A.number_of_rows()) throw ls_error(name,q1,dimension_error_);
  ls_UINT i, j;
  for(i=0;i<n;i++)
  { A[i][i]=(*this)[i][i];
    for(j=i+1;j<n;j++) A[i][j]=A[j][i]=(*this)[i][j];
  }
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::append_column(ls_UINT l)
{ static char *q1="ls_arrayU::append_column";
  if(!dim) throw ls_error(name,q1,object_is_empty_);
  ls_arrayU<T> c(n+l); c.put_array(*this);
  remove(); swap(c);
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::swap(ls_UINT i, ls_UINT k)
{ static char *q1="ls_arrayU::swap";
  if(!(k<n)||!(i<n)) throw ls_error(name,q1,index_error_);
  if(i==k) return *this;
  ls_array1<T> ai(n), ak(n); ls_REAL s;
  get_row(ai,i,i); if(i){ ai.put_array(ai,i); get_column(ai,0,i);}
  get_row(ak,k,k); if(k){ ak.put_array(ak,k); get_column(ak,0,k);}
  s=ai[k], ai[k]=ai[i], ai[i]=s, s=ak[k], ak[k]=ak[i], ak[i]=s;
  if(i){ put_column(ak,0,i); ak.get_array(ak,i);} put_row(ak,i,i);
  if(i){ put_column(ai,0,k); ai.get_array(ai,k);} put_row(ai,k,k);
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::remove(ls_UINT i)
{ static char *q1="ls_arrayU::remove";
  if(!dim) throw ls_error(name,q1,object_is_empty_);
  if(!(i<n)) throw ls_error(name,q1,index_error_);
  swap(0,i);
  T *aa=new T[dim-n]; memcpy(aa,a+n,sizeof(T)*(dim-n));
  delete[]a; dim-=n; n--; a=aa;
  if(i) i--; if(i) swap(0,i);
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::remove()
{ ls_array1<T>::remove(); n=0; return *this;}

template<class T>
const ls_arrayU<T>& ls_arrayU<T>::write_row(ostream &f) const
{ static char *q1="ls_arrayU::write_row";
  if(!n) throw ls_error(q1,name,object_is_empty_);
  T *aa; ls_UINT i, j;
  f<<name<<endl<<"{ "<<dimension_<<" "<<n<<endl;
  for (i=0; i<n; i++, f<<endl)
  { aa=(*this)[i];
    for (j=i; j<n; j++)
      if(aa[j]<0) f << " " << aa[j]; else f << "  " << aa[j];
  }
  f<<"}"<<endl;
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::read_row(istream &f)
{ static char *q1="ls_arrayU::read_row";
  if(dim) throw ls_error(name, q1, object_not_empty_);
  char w[ls_len]; ls_UINT i, j;
  if(*name) while(f>>w&&strcmp(w,name)); else f>>name;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  if(!(f>>w)||strcmp(w,dimension_))
    throw ls_error(name,q1,object_contents_error_);
  if(!(f>>n)||!(f>>w))
    throw ls_error(name,q1,object_contents_error_);
  a=new T[dim=(n*(n+1))/2];
  for (i=0; i<n; i++) for(j=i; j<n; j++) f >>(*this)[i][j];
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  return *this;
}

template<class T>
const ls_arrayU<T>& ls_arrayU<T>::write_column(ostream &f) const
{ static char *q1="ls_array2::write_column";
  if(!n) throw ls_error(q1,name,object_is_empty_);
  ls_UINT i, j; ls_REAL s;
  f <<name<<endl<<"{ "<<dimension_<<" "<<n<<endl;
  for (j=0; j<n; j++) for(i=j; i<j+1; i++)
    if((s=(*this)[i][j])<0) f << " " << s; else f << "  " << s;
  f<<"}"<<endl;
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::read_column(istream &f)
{ static char *q1="ls_arrayU::read_column";
  if(dim) throw ls_error(name, q1, object_not_empty_);
  char w[ls_len]; ls_UINT i, j;
  if(*name) while(f>>w&&strcmp(w,name)); else f>>name;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  if(!(f>>w)||strcmp(w, dimension_))
    throw ls_error(name,q1,object_contents_error_);
  if(!(f>>n)||!(f>>w))
    throw ls_error(name,q1,object_contents_error_);
  a=new T[dim=(n*(n+1))/2];
  for (j=0; j<n; j++)
  { for(i=j; i<j+1; f>>(*this)[i++][j]);}
  if(!(f>>w)||(*w!='}')) throw ls_error(name, q1, size_error_);
  return *this;
}

template<class T>
const ls_arrayU<T>& ls_arrayU<T>::operator>> (char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
	 }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield),
    f.precision(int(sizeof(ls_REAL)*1.5)), write_row(f);
  }
  return *this;
}

template<class T>
ls_arrayU<T>& ls_arrayU<T>::operator<< (char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
