//: $CC -DLS_LIB -c $file
#ifndef LS_SMATRIXC
#define LS_SMATRIXC
#include "ls_names.h"
#include LS_SMATRIX_H

ls_sMatrix::ls_sMatrix(ls_UINT nn, char *nam):
ls_arrayU<ls_REAL>(nn,nam){}

ls_sMatrix::ls_sMatrix(ls_sMatrix &A):ls_arrayU<ls_REAL>(A){}

const ls_sMatrix& ls_sMatrix::operator=(const ls_sMatrix &B)
{ ls_arrayU<ls_REAL>::operator=(B); return *this;}

ls_sMatrix& ls_sMatrix::operator+=(const ls_sMatrix &B)
{ ls_REAL *ba=B.asArray(), *be=ba+dim, *aa=a;
  while(ba<be) *aa+=*ba, aa++, ba++;
  return *this;
}

ls_sMatrix ls_sMatrix::operator-(const ls_sMatrix &B) const
{ ls_sMatrix C; C=*this; C-=B; return C;}

ls_sMatrix& ls_sMatrix::operator-=(const ls_sMatrix &B)
{ ls_REAL *ba=B.asArray(), *be=ba+dim, *aa=a;
  while(ba<be) *aa-=*ba, aa++, ba++;
  return *this;
}

ls_sMatrix ls_sMatrix::operator+(const ls_sMatrix &B) const
{ ls_sMatrix C(n); C=*this; C+=B; return C;}

ls_Vector ls_sMatrix::operator*(const ls_Vector &x) const
{ ls_array1<double> yp(n), ym(n);
  ls_REAL *aa=a, *xx=x.asArray(), *xl;
  double s, sp, sm, *yyp=yp.asArray(), *yym=ym.asArray();
  ls_UINT i, j;
  for (i=0; i<n; yyp[i]+=sp, yym[i]+=sm, xx++, i++)
  { sp=sm=0; if((s=(*aa)*(*xx))<0) sm=s; else sp=s;
    for (xl=xx+1, j=i+1, aa++; j<n; xl++, aa++, j++)
    { if((s=(*aa)*(*xl))<0) sm+=s; else sp+=s;
      if((s=(*aa)*(*xx))<0) yym[j]+=s; else yyp[j]+=s;
    }
  }
  ls_Vector y(n);
  ls_REAL *yy=y.asArray(), *ye=yy+n;
  for(; yy<ye; *yy=(ls_REAL)(*yyp+(*yym)), yy++, yyp++, yym++);
  return y;
}

ls_sMatrix ls_sMatrix::operator*(ls_REAL s) const
{ ls_sMatrix B(n); B=*this; B*=s; return B;}

ls_sMatrix& ls_sMatrix::operator*=(ls_REAL s)
{ for (ls_REAL *aa=a, *ae=aa+dim; aa<ae; *aa*=s, aa++);
  return *this;
}

ls_Matrix ls_sMatrix::operator*(const ls_sMatrix &B) const
{ static char *q1="ls_sMatrix::*";
  if(n!=B.dimension()) throw ls_error(name,q1,size_error_);
  ls_Matrix C(n,n);
  for(ls_UINT i=0; i<n; C.put_row(B*row(i),i), i++);
  return C;
}

ls_UINT ls_sMatrix::solve(ls_Vector &x, ls_Vector &b)
{ b-=(*this)*x;
  ls_Vector p, w; p=b; w=(*this)*p;
  ls_REAL s, bb=b*b, wp=w*p, bbn, eps=x.eps;
  ls_UINT it=0, rep=n*n;
  while(!((bb+wp)<eps)&&(it++<rep))
  { s=bb/wp; b-=s*w; x+=s*p; bbn=b*b; p=b+(bbn/bb)*p;
    w=(*this)*p; wp=w*p; bb=bbn;
  }
  return it;
}

ls_sMatrix ls_Vector::dyad() const
{ ls_sMatrix A(dim);
  ls_REAL *uu, *ue=a, *vv, *aa=A.asArray(); ue+=dim;
  for(uu=a; uu<ue; uu++)
    for (vv=uu; vv<ue; *aa=*uu*(*vv), aa++, vv++);
  return A;
}

ls_sMatrix operator*(ls_REAL x, const ls_sMatrix &B)
{ ls_sMatrix C(B.dimension()); C=B; C*=x; return C;}
#endif
