//: $CC -DLS_LIB -c $file
#ifndef LS_UMATRIXC
#define LS_UMATRIXC
#include "ls_names.h"
#include LS_UMATRIX_H

ls_uMatrix::ls_uMatrix(ls_UINT nn, char *nam):
ls_arrayU<ls_REAL>(nn,nam){}

ls_uMatrix::ls_uMatrix(ls_uMatrix &A):ls_arrayU<ls_REAL>(A){}

const ls_uMatrix& ls_uMatrix::operator=(const ls_uMatrix &B)
{ ls_arrayU<ls_REAL>::operator=(B); return *this;}

ls_uMatrix& ls_uMatrix::operator+=(const ls_uMatrix &B)
{ ls_REAL *ba=B.asArray(), *be=ba+dim, *aa=a;
  while(ba<be) *aa+=*ba, aa++, ba++;
  return *this;
}

ls_uMatrix ls_uMatrix::operator-(const ls_uMatrix &B) const
{ ls_uMatrix C; C=*this; C-=B; return C;}

ls_uMatrix& ls_uMatrix::operator-=(const ls_uMatrix &B)
{ ls_REAL *ba=B.asArray(), *be=ba+dim, *aa=a;
  while(ba<be) *aa-=*ba, aa++, ba++;
  return *this;
}

ls_uMatrix ls_uMatrix::operator+(const ls_uMatrix &B) const
{ ls_uMatrix C(n); C=*this; C+=B; return C;}

ls_Vector ls_uMatrix::operator*(const ls_Vector &x) const
{ ls_Vector y(n);
  ls_REAL *aa=a, *xx=x.asArray(), *xl, *xe=xx+n, *yy=y.asArray();
  double s, sp, sm;
  for (; xx<xe; *yy+=(ls_REAL)(sp+sm), xx++, yy++)
    for (sp=sm=0, xl=xx; xl<xe; xl++, aa++)
      if((s=(*aa)*(*xl))<0) sm+=s; else sp+=s;
  return y;
}

ls_uMatrix ls_uMatrix::operator*(ls_REAL s) const
{ ls_uMatrix B(n); B=*this; B*=s; return B;}

ls_uMatrix& ls_uMatrix::operator*=(ls_REAL s)
{ for (ls_REAL *aa=a, *ae=aa+dim; aa<ae; *aa*=s, aa++);
  return *this;
}

ls_Matrix ls_uMatrix::operator*(const ls_uMatrix &B) const
{ static char *q1="ls_uMatrix::*";
  if(n!=B.dimension()) throw ls_error(name,q1,size_error_);
  ls_Matrix C(n,n);
  for(ls_UINT i=0; i<n; C.put_row(B*row(i),i), i++);
  return C;
}

ls_UINT ls_uMatrix::solve(ls_Vector &x, ls_Vector &b)
{ b-=(*this)*x; ls_Vector p; p=b*(*this);
  ls_REAL s, bb=b*b, bbn, pp=p*p, eps=x.eps;
  ls_UINT it=0, rep=n*n;
  while (!((bb+pp)<eps)&&(it++<rep))
  { s=bb/pp, x+=s*p;
    b-=(s*(*this)*p);
    bbn=b*b;
    p=b*(*this)+(bbn/bb)*p,  pp=p*p, bb=bbn;
  }
  return it;
}

void ls_uMatrix::backward(ls_Vector &x, ls_Vector &b)
{ ls_UINT i,j; ls_REAL s, *aa;
  for(i=n; i--; x[i]=(b[i]-s)/aa[i])
    for(s=0, aa=(*this)[i], j=i; ++j<n; s+=aa[j]*x[j]);
}

ls_uMatrix operator*(ls_REAL x, const ls_uMatrix &B)
{ ls_uMatrix C(B.dimension()); C=B; C*=x; return C;}

ls_Vector operator*(const ls_Vector &x, const ls_uMatrix &B)
{ ls_UINT n=x.dimension();
  ls_Vector y(n);
  ls_REAL *aa=B.asArray(), *xx=x.asArray(), *ya=y.asArray(), 
          *yy, *ye=ya+n, *yl;
  for (yy=ya; yy<ye; yy++, xx++)
    for (yl=yy; yl<ye; *yl+=*aa*(*xx), aa++, *yl++)
  return y;
}
#endif
